# coding=utf-8
import torch
import os
import datetime
import unicodedata
import json

class InputFeatures(object):
    def __init__(self, input_id, label_id, input_mask):
        self.input_id = input_id
        self.label_id = label_id
        self.input_mask = input_mask


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = {}
    index = 0
    with open(vocab_file, "r", encoding="utf-8") as reader:
        while True:
            token = reader.readline()
            if not token:
                break
            token = token.strip()
            vocab[token] = index
            index += 1
    return vocab


def read_corpus(path, max_length, label_dic, vocab):
    """
    :param path:数据文件路径
    :param max_length: 最大长度
    :param label_dic: 标签字典
    :return:
    """
    # file = open(path, encoding='utf-8')
    # content = file.readlines()
    # file.close()
    result = []
    # with open(path, 'r', encoding = 'utf-8') as f:

    tjson=Tjson(file_path=path)

    for line in tjson.load():
        # text, label = line.strip().split('|||')
        
        # tokens = text.split()
        # label = label.split()
        # print(line)
        tokens=line['text']
        label=line['label']

        if len(tokens) > max_length-2:
            tokens = tokens[0:(max_length-2)]
            label = label[0:(max_length-2)]
        tokens_f =['[CLS]'] + tokens + ['[SEP]']
        label_f = ["<start>"] + label + ['<eos>']
        input_ids = [int(vocab[i]) if i in vocab else int(vocab['[UNK]']) for i in tokens_f]
        label_ids = [label_dic[i] for i in label_f]
        input_mask = [1] * len(input_ids)
        while len(input_ids) < max_length:
            input_ids.append(0)
            input_mask.append(0)
            label_ids.append(label_dic['<pad>'])
        assert len(input_ids) == max_length
        assert len(input_mask) == max_length
        assert len(label_ids) == max_length
        feature = InputFeatures(input_id=input_ids, input_mask=input_mask, label_id=label_ids)
        result.append(feature)
    return result

class InputPreFeatures(object):
    def __init__(self, input_id, input_mask):
        self.input_id = input_id
        # self.label_id = label_id
        self.input_mask = input_mask

def build_input(content, max_length, vocab):
    """
    :param content: 输入句子列表
    :param max_length: 最大长度
    :return:
    """
    # file = open(path, encoding='utf-8')
    # content = file.readlines()
    # file.close()
    result = []
    for line in content:
        # text, label = line.strip().split('|||')
        tokens = line.split()
        # label = label.split()
        if len(tokens) > max_length-2:
            tokens = tokens[0:(max_length-2)]
            # label = label[0:(max_length-2)]
        # print(tokens)
        tokens_f =['[CLS]'] + tokens + ['[SEP]']
        # print('tokens_f',tokens_f)
        # label_f = ["<start>"] + label + ['<eos>']
        input_ids = [int(vocab[i]) if i in vocab else int(vocab['[UNK]']) for i in tokens_f]
        # label_ids = [label_dic[i] for i in label_f]
        input_mask = [1] * len(input_ids)
        while len(input_ids) < max_length:
            input_ids.append(0)
            input_mask.append(0)
            # label_ids.append(label_dic['<pad>'])
        assert len(input_ids) == max_length
        assert len(input_mask) == max_length
        # assert len(label_ids) == max_length
        feature = InputPreFeatures(input_id=input_ids, input_mask=input_mask)
        result.append(feature)
    return result



def save_model(model, epoch, path='result', **kwargs):
    """
    默认保留所有模型
    :param model: 模型
    :param path: 保存路径
    :param loss: 校验损失
    :param last_loss: 最佳epoch损失
    :param kwargs: every_epoch or best_epoch
    :return:
    """
    if not os.path.exists(path):
        os.mkdir(path)
    if kwargs.get('name', None) is None:
        # cur_time = datetime.datetime.now().strftime('%Y-%m-%d#%H:%M:%S')
        # name = cur_time + '--epoch:{}'.format(epoch)
        name="pytorch_model.bin"
        full_name = os.path.join(path, name)
        torch.save(model.state_dict(), full_name)
        # print('Saved model at epoch {} successfully'.format(epoch))
        with open('{}/checkpoint'.format(path), 'w') as file:
            file.write(name)
            # print('Write to checkpoint')


def load_model(model, path='result', **kwargs):
    if kwargs.get('name', None) is None:
        with open('{}/checkpoint'.format(path)) as file:
            content = file.read().strip()
            name = os.path.join(path, content)
    else:
        name=kwargs['name']
        # name = os.path.join(path,name)
        name = os.path.join(name)
    # print('name',name)
    model.load_state_dict(torch.load(name, map_location=lambda storage, loc: storage))
    # print('load model {} successfully'.format(name))
    return model



class Tjson:
  """
  处理json信息函数
  """
  def __init__(self,file_path="data.json"):
    self.file_path=file_path
  def save(self,data):
    """
    保存数据函数
    逐行写入
    >>> data=[{'a':'ess'}]
    """
    with open(self.file_path, 'a+', encoding='utf-8') as f:
      for item in data:
        line = json.dumps(item, ensure_ascii=False)
        f.write(line+'\n')


  def load(self):
    """
    加载数据

    """
    with open(self.file_path, 'r', encoding = 'utf-8') as f:
        for line in f:
            data=json.loads(line[:-1])
            yield data
"""
#使用
data=[{'a':'ess'}]
Json().save(data)
print(Json().load())


"""


# coding=utf-8
import torch

class Config(object):
    def __init__(self):
        self.label_file = './data/tag.txt'
        self.train_file = './data/train.json'
        self.dev_file = './data/dev.json'
        self.test_file = './data/test.txt'
        # self.vocab = './data/bert/vocab.txt'
        self.vocab = './data/albert/vocab.txt'
        self.max_length = 50
        # device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        if torch.cuda.is_available():
            self.use_cuda = True
        else:
            self.use_cuda = False
              
        self.gpu = 0
        self.batch_size = 10
        self.bert_path = './data/bert'
        self.albert_path = './data/albert'
        self.rnn_hidden = 500
        self.bert_embedding = 768
        self.albert_embedding= 768
        self.dropout1 = 0.5
        self.dropout_ratio = 0.5
        self.rnn_layer = 1
        self.lr = 0.0001
        self.lr_decay = 0.00001
        self.weight_decay = 0.00005
        self.checkpoint = 'result/'
        self.optim = 'Adam'
        self.load_model = False
        self.load_path = None
        self.base_epoch = 10

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
    def update_json(self, args):
        """
        修改json版本数据
        """
        for k, v in args.items():
            setattr(self, k, v)
    def __str__(self):

        return '\n'.join(['%s:%s' % item for item in self.__dict__.items()])


if __name__ == '__main__':

    con = Config()
    con.update(gpu=8)
    print(con.gpu)
    print(con)



# coding=utf-8
import torch
import torch.nn as nn
from torch.autograd import Variable
from config import Config
from model import ALBERT_LSTM_CRF
import torch.optim as optim
from utils import load_vocab, read_corpus, load_model, save_model,build_input,Tjson
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader
import fire




class Pre:
    """
    执行预测操作
    content=["柯基犬是个狗子"]
    P=Pre()
    result=P.pre(content)
    print(result)  

    >>>[('柯基犬真是和牛逼', [{'type': '实体', 'words': ['柯', '基', '犬']}])]  
    """
    def __init__(self):
        self.args={
            "load_path":"result/pytorch_model.bin",
            "load_model":True,
            "use_cuda":False,
        }
        
        config = Config()
        config.update_json(self.args)
        self.config=config
        print( self.config)
        self.load_model()

    def load_model(self):
        """
        加载模型
        """

        # config = Config()
        # config.update_json(self.args)
        config=self.config
        # print('当前设置为:\n', config)
        if config.use_cuda:
            torch.cuda.set_device(config.gpu)
        # print('loading corpus')
        vocab = load_vocab(config.vocab)
        label_dic = load_vocab(config.label_file)
        tagset_size = len(label_dic)
        model = ALBERT_LSTM_CRF(config.albert_path, tagset_size, config.albert_embedding, config.rnn_hidden, config.rnn_layer, dropout_ratio=config.dropout_ratio, dropout1=config.dropout1, use_cuda=config.use_cuda)
        if config.load_model:
            assert config.load_path is not None
            # 
            model = load_model(model, name=config.load_path)
        if config.use_cuda:
            model.cuda()
        # return model, vocab, label_dic
        self.model, self.vocab, self.label_dic=model, vocab, label_dic

    def build_input(self,content_array):
        """
        基于文本列表构建训练数据集
        """
        config=self.config
        model, vocab, label_dic=self.model, self.vocab, self.label_dic
        new_content_array=[]
        content_dict={}
        for i, content in enumerate (content_array):
            new_content_array.append(" ".join(list(content)))
            # content_dict[i]=content
        # print(content)
        # print(content_dict)
        input_data = build_input(content=new_content_array, max_length=self.config.max_length, vocab=self.vocab)
        # print(input_data)
        input_ids = torch.LongTensor([temp.input_id for temp in input_data])
        input_masks = torch.LongTensor([temp.input_mask for temp in input_data])

        input_dataset = TensorDataset(input_ids, input_masks)
        input_loader = DataLoader(input_dataset, shuffle=True, batch_size=self.config.batch_size)
        return input_loader

    def pre(self,content_array=[]):
        """
        执行预测
        输入多个句子即可
        """

        
        config=self.config
        model, vocab, label_dic=self.model, self.vocab, self.label_dic

        content_dict={}
        for i, content in enumerate (content_array):
            # new_content_array.append(" ".join(list(content)))
            content_dict[i]=content

        output=[]
        for i, batch in enumerate(self.build_input(content_array)):
            inputs, masks = batch
            # print('inputs',inputs)
            inputs, masks= Variable(inputs), Variable(masks)
            # print("masks",masks)
            if self.config.use_cuda:
                inputs, masks = inputs.cuda(), masks.cuda()
            feats = model(inputs)
            # print("feats",feats)
            path_score, best_path = model.crf(feats, masks.bool())
            # print("feats",path_score, best_path)
            for item in best_path.numpy():
                # print(item.tolist())
                words=[]
                for n,id in enumerate( item.tolist()):
                    word_id=inputs.numpy().tolist()[0][n]
                    l=list(label_dic)[id]
                    w=list(self.vocab)[word_id]
                    words.append((w,l))

                # print('words',words)
                wd={"type":None,"words":[]}
                wd_list=[]
                for w,l in words:
                    # print(wd)
                    if l.startswith("B-"): 
                        wd["type"]=l.split("B-")[1]
                        wd["words"].append(w)
                    elif l.startswith("M-"):
                        wd["type"]=l.split("M-")[1]
                        wd["words"].append(w)
                    elif l.startswith("E-"):
                        wd["type"]=l.split("E-")[1]
                        wd["words"].append(w)
                        wd_list.append(wd)
                        wd={"type":None,"words":[]}
                    elif l.startswith("S-"):
                        wd["type"]=l.split("S-")[1]
                        wd["words"].append(w)
                        wd_list.append(wd)
                        wd={"type":None,"words":[]}
                # print('wd_list',wd_list)
                # print(i)
                output.append((content_dict[i],wd_list))
        return output






def pre_one(**kwargs):
    # content="柯基犬性格活泼可爱，但是饲养柯基犬会有着六个坏处，你还敢饲养柯基犬吗？"
    # print(**kwargs)
    P=Pre()
    result=P.pre([kwargs['text']])
    print(result)
    




if __name__ == '__main__':
    fire.Fire()
    # test()










